"""Test libopenmm."""
import importlib
import sys

import pytest
from openmm.app import pdbxfile

from taurenmd.libs import libopenmm

from . import toptest, toptest_cif, trajtest


def test_simtk():
    """Test simtk is installed."""
    assert libopenmm.SIMTK


def test_attempt_load_cif_SIMTK_1():
    """Test attempt load CIF."""
    mol = libopenmm.attempt_to_load_top_from_simtk(toptest_cif)
    assert isinstance(mol, pdbxfile.PDBxFile)


def test_attempt_load_cif_SIMTK_2():
    """Test attempt load CIF."""
    with pytest.raises(ValueError):
        libopenmm.attempt_to_load_top_from_simtk(toptest)


def test_simtk_import_error():
    """Test import error message."""
    with pytest.raises(SystemExit) as err:
        libopenmm._log_simtkimport_error()
    assert err.type == SystemExit
    assert err.value.code == 0


def test_without_simtk():
    """Test without simtk installed."""
    sys.modules['openmm.app.pdbxfile'] = None
    importlib.reload(libopenmm)
    with pytest.raises(SystemExit) as err:
        libopenmm.attempt_to_load_top_from_simtk(toptest_cif)
    assert err.type == SystemExit
    assert err.value.code == 0


def test_load_traj_cif_import_error():
    """Test loadtraj_cif import error as if simtk was not installed."""
    sys.modules['openmm.app.pdbxfile'] = None
    from taurenmd.libs import libmdt
    importlib.reload(libmdt)
    with pytest.raises(SystemExit) as err:
        libmdt.load_traj(toptest_cif, trajtest)
    assert err.type == SystemExit
    assert err.value.code == 0



"""Test libmdt."""
import mdtraj as md

from taurenmd.libs import libmdt

from . import toptest, toptest_cif, trajtest


def test_load_traj_pdb():
    """Test loading traj."""
    traj = libmdt.load_traj(toptest, trajtest)
    assert isinstance(traj, md.Trajectory)
    assert len(traj) == 10


def test_load_traj_pdb_2():
    """Test loading traj."""
    traj = libmdt.load_traj(toptest, [trajtest, trajtest])
    assert isinstance(traj, md.Trajectory)
    assert len(traj) == 20


def test_load_traj_pdb_3():
    """Test loading traj."""
    traj = libmdt.load_traj(toptest, [trajtest, trajtest], insort=True)
    assert isinstance(traj, md.Trajectory)
    assert len(traj) == 20


def test_load_traj_cif():
    """Test loading traj."""
    assert toptest_cif.suffix == '.cif'
    traj = libmdt.load_traj(toptest_cif, trajtest)
    assert isinstance(traj, md.Trajectory)
    assert len(traj) == 10



"""
Functions that wrap around `OpenMM library`_.

Read our `citing documentation`_ to understand how to cite multiple
libraries.

.. _citing documentation: https://taurenmd.readthedocs.io/en/latest/citing.html
.. _OpenMM library: http://openmm.org/
"""
import sys


try:
    from openmm.app.pdbxfile import PDBxFile
    SIMTK = True
except ImportError:
    SIMTK = False


from taurenmd import Path
from taurenmd import core as tcore
from taurenmd import log
from taurenmd.libs import libcli
from taurenmd.logger import S, T


@libcli.add_reference(tcore.ref_openmm)
def attempt_to_load_top_from_simtk(topology):
    """
    Load topology from SIMTK.

    Parameters
    ----------
    topology : str or Path

    Returns
    -------
    topology from mdtraj.Topology.from_openmm`

    Raises
    ------
    Dependency error from :func:`_log_simtkimport_error`, program
    halts.
    """
    topp = Path(topology)

    if topp.suffix == '.cif' and SIMTK:
        mol = PDBxFile(topp.str())
        return mol

    elif topp.suffix == '.cif' and not SIMTK:
        _log_simtkimport_error()

    else:
        raise ValueError(f'`topology` suffix is not CIF: {topp.suffix!r}')


def _log_simtkimport_error():
    msg = (
        "To use .cif files as Topologies taurenmd requires OpenMM, "
        "which is currently not installed. "
        "Please visit our Installation instruction at "
        "https://taurenmd.readthedocs.io/"
        )
    log.error(T('Dependency Error'))
    log.error(S(msg))
    sys.exit(0)



"""
Certify the developer has input all requirements for PR.

Situations tested:

* additions are reported in CHANGELOG.rst
"""
from pathlib import Path

folder = Path(__file__).resolve().parents[1]
changelog = Path(folder, 'CHANGELOG.rst')

with open(changelog, 'r') as fin:
    for line in fin:
        if line.startswith('v'):
            raise ValueError(
                'Please add a summary of your additions to docs/CHANGELOG.rst. '
                'As described in: https://python-project-skeleton.readthedocs.io'
                '/en/latest/contributing.html#update-changelog.'
                )
        elif line.startswith('*'):
            break



"""
# Decompose Eurler angle rotations of a selection.

*Calculate the Roll, Pitch and Yaw angles along the trajectory.*

Read further on roll, pitch and yaw angles (Euler Angles) -
[wikipedia](https://en.wikipedia.org/wiki/Euler_angles).

Here we decompose these movements around the three different axis
centered at an origin.

## Algorithm

Given a selection of three regions, ``selection A``, ``selection B``
and ``selection C``:

1. Centers the system to the three selections center of geometry for
every frame, this is called the *origin*,

2. Calculates a plane given by the center of geometries of the three
selections, plane ABC,

3. Defines the vector OA that goes from the origin to the center of
geometry of ``selection A``, this represents the Yaw axis.

4. Defines the normal vector to the plane ABC (ABCn), this represents the
Roll axis,

5. Defines the cross product betwee vectors OA and ABCn (AONn), this is the
Pitch axis,

6. Calculates this axis of reference for every frame

Calculating the angles:

Angles represent the right hand rotation around an axis of the sistem in
a i-frame compared to the reference frame.

### Roll

The roll angle is given by the torsion angle between OA, origin, ABCn,
and OAi (in frame), displaced along ABCn.

### Pitch

The pitch angle is by the torsion angle between ABCn, origin, AONn, and
ABCni (in frame), displaced along AONn.

### Yaw

The yaw angle is given by the torsion angle between the AONn, origin,
OA, and AONni (in frame), displaced along OA.

## Examples

In the case of an homotrimer, define the axis and the origin on the trimers:

    taurenmd rorations -z 'segid A' 'segid B' 'segid C' -x rotations.csv

## References

"""  # noqa: E501
import argparse
import functools

import numpy as np

from taurenmd import _BANNER, Path
from taurenmd import core as tcore
from taurenmd import log
from taurenmd.libs import libcalc, libcli, libio, libmda, libplot  # noqa: F401
from taurenmd.logger import S, T


__author__ = 'Joao M.C. Teixeira'
__email__ = 'joaomcteixeira@gmail.com'
__maintainer__ = 'Joao M.C. Teixeira'
__credits__ = ['Joao M.C. Teixeira']
__status__ = 'Production'

__doc__ += (
    f'{tcore.ref_mda}'
    f'{tcore.ref_mda_selection}'
    )

_help = 'Calculates angular rotations across axes.'
_name = 'rotations'

ap = libcli.CustomParser(
    description=_BANNER + __doc__,
    formatter_class=argparse.RawDescriptionHelpFormatter,
    )

libcli.add_version_arg(ap)
libcli.add_topology_arg(ap)
libcli.add_trajectories_arg(ap)
libcli.add_insort_arg(ap)
libcli.add_plane_selection_arg(ap)
libcli.add_angle_unit_arg(ap)
libcli.add_reference_frame_arg(ap)
libcli.add_slice_arg(ap)
libcli.add_data_export_arg(ap)
libcli.add_plot_arg(ap)


def _ap():
    return ap


def main(
        topology,
        trajectories,
        plane_selection,
        insort=False,
        aunit='degrees',
        ref_frame=0,
        start=None,
        stop=None,
        step=None,
        export=False,
        plot=False,
        plotvars=None,
        **kwargs,
        ):
    """Execute main client logic."""
    log.info(T('starting'))

    topology = Path(topology)
    trajectories = list(map(Path, trajectories))
    u = libmda.load_universe(topology, *trajectories, insort=insort)

    log.info(T('slicing'))
    fSlice = libio.frame_slice(start=start, stop=stop, step=step)

    log.info(T('computing planes'))
    log.info(S('for reference frame: {}', ref_frame))

    ABC_selections = plane_selection  # variable renaming
    origin_selection = ' or '.join(ABC_selections)
    log.info(S('for selection: {!r}', origin_selection))

    # p stands for point
    # atomG stands for MDAnalysis AtomGroup
    pABC_atomG = u.select_atoms(origin_selection)
    pA_atomG = u.select_atoms(ABC_selections[0])
    pB_atomG = u.select_atoms(ABC_selections[1])
    pC_atomG = u.select_atoms(ABC_selections[2])

    u.trajectory[ref_frame]

    # defining the center of reference
    opABC_cog = pABC_atomG.center_of_geometry()
    log.info(S('Original pABC center of geometry: {}', opABC_cog))

    log.info(S(
        'transferring all coordinates of '
        'reference frame to the origin (0, 0, 0)'
        ))
    pABC_atomG.positions = pABC_atomG.positions - opABC_cog
    pABC_cog = pABC_atomG.center_of_geometry().copy()  # this should be zero
    log.info(S('New origin Center of Geometry'))
    log.info(S('pABC_cog: {}', pABC_cog))

    log.info(T('defining the reference axes'))
    # vector from the origin (0, 0, 0) to cog of each selection
    pA_cog = pA_atomG.center_of_geometry().copy()
    pB_cog = pB_atomG.center_of_geometry().copy()
    pC_cog = pC_atomG.center_of_geometry().copy()
    log.info(S('plane points definition:'))
    log.info(S('pA: {}', pA_cog))
    log.info(S('pB: {}', pB_cog))
    log.info(S('pC: {}', pC_cog))

    ref_plane_normal = libcalc.calc_plane_normal(pABC_cog, pA_cog, pB_cog)
    log.info(S('Normal vector to the reference plane: {}', ref_plane_normal))

    ref_plane_cross = np.cross(pA_cog, ref_plane_normal)

    log.info(S('Cross vector to pA and Normal vector: {}', ref_plane_cross))

    log.info(T('Calculating tilt angles'))
    total_frames = len(u.trajectory[fSlice])
    log.info(S('for {} frames', total_frames))
    sizet = (total_frames, 3)
    roll_vectors = np.empty(sizet, dtype=np.float64)
    pitch_vectors = np.empty(sizet, dtype=np.float64)
    yaw_vectors = np.empty(sizet, dtype=np.float64)

    with libcli.ProgressBar(total_frames, suffix='frames') as pb:
        for i, _ts in enumerate(u.trajectory[fSlice]):

            pABC_cog_ts = pABC_atomG.center_of_geometry().copy()
            pABC_atomG.positions = pABC_atomG.positions - pABC_cog_ts

            ts_positions = pABC_atomG.positions.copy()

            # get roll
            pABC_atomG.positions = ts_positions + ref_plane_normal
            roll_vectors[i, :] = pA_atomG.center_of_geometry().copy()

            # get pitch
            pABC_atomG.positions = ts_positions + ref_plane_cross
            pitch_vectors[i, :] = libcalc.calc_plane_normal(
                ref_plane_cross,
                pA_atomG.center_of_geometry(),
                pB_atomG.center_of_geometry(),
                )

            # get yaw
            pABC_atomG.positions = ts_positions + pA_cog
            pA_cog_yaw_ts = pA_atomG.center_of_geometry().copy()
            normalv = libcalc.calc_plane_normal(
                pA_cog,
                pA_cog_yaw_ts,
                pB_atomG.center_of_geometry(),
                )
            yaw_vectors[i, :] = np.cross(pA_cog_yaw_ts, normalv)
            pABC_atomG.positions = ts_positions

            pb.increment()

    log.info(S('done'))

    roll_torsion = libcalc.torsion_set(
        pA_cog,
        pABC_cog,
        ref_plane_normal,
        roll_vectors,
        )

    pitch_torsion = libcalc.torsion_set(
        ref_plane_normal,
        pABC_cog,
        ref_plane_cross,
        pitch_vectors,
        )

    yaw_torsion = libcalc.torsion_set(
        ref_plane_cross,
        pABC_cog,
        pA_cog,
        yaw_vectors,
        )

    if aunit == 'degrees':
        roll_torsion, pitch_torsion, yaw_torsion = \
            [np.degrees(a) for a in [roll_torsion, pitch_torsion, yaw_torsion]]

    if export:
        libio.export_data_to_file(
            list(range(len(u.trajectory))[fSlice]),
            roll_torsion, pitch_torsion, yaw_torsion,
            fname=export,
            header=(
                '# Topology: {}\n'
                '# Trajectories: {}\n'
                '# Plane Selection: {}\n'
                '# frame,roll,pitch,yaw [in {}]\n'
                ).format(
                    topology.resolve().str(),
                    ', '.join(t.resolve().str() for t in trajectories),
                    origin_selection,
                    aunit,
                    ),
            )

    if plot:
        plotvars = plotvars or dict()
        plotvars.setdefault('labels', ['roll', 'pitch', 'yaw'])

        libplot.param(
            list(range(len(u.trajectory))[fSlice]),
            [roll_torsion, pitch_torsion, yaw_torsion],
            **plotvars,
            )

    return


maincli = functools.partial(libcli.maincli, ap, main)


if __name__ == '__main__':
    maincli()



"""
Functions that wrap around `MDTraj library`_.

Functions contained in this module operate with MDTraj functionalities,
wither by using MDTraj to access Molecular Dynamics data or by receiving
MDTraj data structures and parsing them in some way.

`Simtk OpenMM`_ is also used in some functions.

Read our `citing documentation`_ to understand how to cite multiple
libraries.

.. _MDTraj library: http://mdtraj.org/1.9.3/index.html
.. _citing documentation: https://taurenmd.readthedocs.io/en/latest/citing.html
.. _Simtk OpenMM: http://openmm.org/
"""
import os

import mdtraj

from taurenmd import Path
from taurenmd import core as tcore
from taurenmd import log
from taurenmd.libs import libcli, libio, libopenmm
from taurenmd.logger import S, T


@libcli.add_reference(tcore.ref_mdt)
def load_traj(topology, trajectories, insort=False):
    """
    Load trajectory with `MDTraj <http://mdtraj.org/1.9.3/index.html>`_.

    Uses `mdtraj.load <http://mdtraj.org/1.9.3/api/generated/mdtraj.load.html?highlight=load#mdtraj.load>`_.

    Example
    -------

        >>> libmdt.load_traj('bigtopology.cif', 'trajectory.dcd')

    Parameters
    ----------
    topology : str or Path or list
        Path to the topology file. Accepts MDTraj compatible `topology files <http://mdtraj.org/1.9.3/load_functions.html#trajectory-reference>`_. mmCIF format is loaded using `OpenMM <http://mdtraj.org/1.9.3/api/generated/mdtraj.Topology.html?highlight=from_openmm#mdtraj.Topology.from_openmm>`_.

    trajectory : str or Path
        Path to the trajectory file. Accepts MDTraj compatible `files <http://mdtraj.org/1.9.3/load_functions.html#trajectory-reference>`_

    Returns
    -------
    MDTraj trajectory
        `Trajectory object <http://mdtraj.org/1.9.3/api/generated/mdtraj.Trajectory.html#mdtraj-trajectory>`_.
    """  # noqa: E501
    if insort:
        trajectories = libio.sort_numbered_input(*trajectories)

    try:
        # just in case Paths arrive
        trajs = [os.fspath(t) for t in trajectories]
    except TypeError:
        trajs = os.fspath(trajectories)

    libio.report_input(topology, trajs)

    if Path(topology).suffix == '.cif':
        _top = libopenmm.attempt_to_load_top_from_simtk(topology)
        top = mdtraj.Topology.from_openmm(_top.topology)
    else:
        top = Path(topology).str()

    mdtrajectory = mdtraj.load(trajs, top=top)

    return mdtrajectory


@libcli.add_reference(tcore.ref_mdt)
def imagemol_protocol1(traj):
    """Attempt to image molecules acting on the whole traj."""
    log.info(T('running reimage protocol #1'))
    log.info(S('finding molecules'))

    mols = traj.top.find_molecules()
    log.info(S('done'))

    log.info(T('reimaging'))
    reimaged = traj.image_molecules(
        inplace=False,
        anchor_molecules=mols[:1],
        other_molecules=mols[1:],
        )
    log.info(S('done'))
    return reimaged


@libcli.add_reference(tcore.ref_mdt)
def imagemol_protocol2(traj):
    """Attempt to image molecules frame by frame."""
    reimaged = []
    for frame in range(len(traj)):
        log.info(S('reimaging frame: {}', frame))

        mols = traj[frame].top.find_molecules()

        reimaged.append(
            traj[frame].image_molecules(
                inplace=False,
                anchor_molecules=mols[:1],
                other_molecules=mols[1:],
                )
            )

    log.info(S('concatenating traj frames'))
    # http://mdtraj.org/1.9.3/api/generated/mdtraj.join.html#mdtraj.join
    reimaged_traj = reimaged[0].join(reimaged[1:])

    return reimaged_traj



"""
Calculate parameters from Molecular Dynamics data.

This module contains functions to calculate MD parameters such as:

#. RMSDs
#. RMSFs
#. plane angle variation
#. axes rotation decomposition

It contains also other functions that help on the calculation of the
desirables. Those functions are also avaible for independent use.

This library contains functions that operate on different Molecular
Dynamics data types. When special data types (MD analysis libraries)
are used, a prefix to the function name is used, and its docstring
explicitly refers to it.

When using these functions, you should always cite taurenmd together
with the other library(ies) used. `Read our citing reference page`_.

.. _Read our citing reference page: https://taurenmd.readthedocs.io/en/latest/citing.html
"""  # noqa: E501
import math

import numpy as np
from MDAnalysis.analysis.rms import RMSD as mdaRMSD
from MDAnalysis.analysis.rms import RMSF as mdaRMSF

from taurenmd import core as tcore
from taurenmd import log
from taurenmd.libs import libcli, libio
from taurenmd.logger import S, T


@libcli.add_reference(tcore.ref_mda)
def mda_rmsd(
        universe,
        frame_slice=None,
        selection='all',
        ref_frame=0,
        ):
    """
    Calculate RMSDs observed for a selection.

    Uses `MDAnalysis RMSD <https://www.mdanalysis.org/docs/documentation_pages/analysis/rms.html?highlight=rmsd#MDAnalysis.analysis.rms.RMSD>`_.

    Example
    -------

    * Calculate RMSDs observed for the whole system along the whole trajectory.

        >>> mda_rmsd(universe)

    * Calculate the RMSDs observed for selection `segid A` for every 10 frames.

        >>> mda_rmsd(universe, slice(0, None, 10), selection='segid A')

    Parameters
    ----------
    MDAnalysis Universe
        The MDAnalysis `universe <https://www.mdanalysis.org/docs/documentation_pages/core/universe.html?highlight=universe#core-object-universe-mdanalysis-core-universe>`_.

    frame_slice : any, optional
        The frames in the trajectory to consider.
        If ``None`` considers all frames.
        Accepts any argument that
        :py:func:`taurenmd.libs.libio.evaluate_to_slice` can receive.
        Defaults to ``None``.

    selection : str, optional
        The selection upon which calculate the RMSDs.
        Defaults to ``'all'``.

    ref_frames : int, optional
        The reference frame against which calculate the RMSDs.
        Defaults to ``0``.

    Returns
    -------
    Numpy Array
        The array containing the calculated RMSDs of shape (N,), where
        N is the number of frames.

    Raises
    ------
    ValueError
        If ``frame_slice`` is not ``None`` or ``slice`` object.

    MDAnalysis Exceptions
        Any exceptions that could come from MDAnalysis RMSF computation.
    """  # noqa: E501
    log.info(T('Calculating RMSDs'))

    frame_slice = libio.evaluate_to_slice(value=frame_slice)

    log.info(S('for selection: {}', selection))
    log.info(S('for trajectory slice: {}', frame_slice))

    R = mdaRMSD(
        universe,
        universe,
        select=selection,
        groupselection=None,
        ref_frame=ref_frame,
        verbose=False,
        )

    R.run(verbose=False)

    # rmsds[:, ii] = R.rmsd[:, 2][self._fslicer]

    return R.rmsd[frame_slice, 2]


@libcli.add_reference(tcore.ref_mda)
def mda_rmsf(
        atom_group,
        frame_slice=None,
        ):
    """
    Calculate RMSFs.

    Uses `MDAnalysis RMSF <https://www.mdanalysis.org/docs/documentation_pages/analysis/rms.html?highlight=rmsd#MDAnalysis.analysis.rms.RMSF>`_.

    Parameters
    ----------
    atom_group : MDAnalysis Atom Group.
       `MDAnalysis Atom group <https://www.mdanalysis.org/docs/documentation_pages/core/groups.html?highlight=atom%20group#MDAnalysis.core.groups.AtomGroup>`_.

    frame_slice : any, optional
        Any argument that :py:func:`taurenmd.libs.libio.evaluate_to_slice` can receive.
        Defaults to ``None``, considers all frames.

    Returns
    -------
    Numpy Array
        With the calculated RMSFs values, of shape (N,) where N are the
        frames sliced from ``frame_slice``.

    Raises
    ------
    MDAnalysis Exceptions
        Any exceptions that could come from MDAnalysis RMSF computation.
    """  # noqa: E501
    log.info(T('Calculating RMSFs'))

    frame_slice = libio.evaluate_to_slice(value=frame_slice)

    log.info(S('for trajectory slice: {}', frame_slice))

    R = mdaRMSF(
        atom_group,
        start=frame_slice.start,
        stop=frame_slice.stop,
        step=frame_slice.step,
        verbose=False,
        )

    R.run()

    return R.rmsf


def calc_plane_normal(p1, p2, p3):
    """
    Calculate the normal vector for the (p1, p2, p3) plane.

    Given 3 XYZ space coordinate points, calculates the normal vector
    of the plane defined by those points.

    Parameters
    ----------
    p1, p2, p3 : numpy.ndarray of shape (3,)
        The three 3D coordinate points that define the plane.

    Returns
    -------
    Numpy array of shape (3,)
        The normal vector to the (p1, p2, p3) plane. This vector
        is **NOT** an unitary vector.
    """
    v1 = p2 - p1
    v2 = p3 - p1
    return np.cross(v1, v2)  # normal to the plane


def calc_plane_eq(p1, p2, p3):
    """
    Calculate equation that defines the (p1, p2, p3) plane.

    `Further reading <http://kitchingroup.cheme.cmu.edu/blog/2015/01/18/Equation-of-a-plane-through-three-points/>`_.

    Parameters
    ----------
    p1, p2, p3 : numpy.ndarray of shape (3,)
        The three 3D coordinate points that define the plane.

    Returns
    -------
    tuple of length 4
        The four parameters (a, b, c, d) that defined the plane equation:

        .. math::

            ax + by + cz = d
    """  # noqa: E501
    cp = calc_plane_normal(p1, p2, p3)
    a, b, c = cp
    d = np.dot(cp, p3)
    return a, b, c, d


def calc_planes_angle(a1, b1, c1, a2, b2, c2, aunit='radians'):
    """
    Calculate the angle between two planes.

    Plane 1 is defined by a1, b1, c1 plane parameters,
    plane 2 is defined by a2, b2, c2, where:

    .. math::

        a1*x + b1*y + c1*z + d = 0

        a2*x + b2*y + c2*z + d = 0

    `Read further <from: https://www.geeksforgeeks.org/angle-between-two-planes-in-3d/>`_.

    Parameters
    ----------
    a1, b1, c1, a2, b2, c2 : float
        Plane parameters

    angle : str, optional
        ``degrees`` returns angle quantity in degrees, else returns
        in radians.

    Returns
    -------
    float
        The angle between plane 1 and plane 2.
    """  # noqa: E501
    d = (a1 * a2 + b1 * b2 + c1 * c2)
    e1 = math.sqrt(a1 * a1 + b1 * b1 + c1 * c1)
    e2 = math.sqrt(a2 * a2 + b2 * b2 + c2 * c2)
    d = d / (e1 * e2)
    angle = math.acos(d)
    if aunit == 'degrees':
        return math.degrees(angle)
    else:
        return angle


def calc_torsion_angles(coords):
    """
    Calculate torsion angles from sequential coordinates.

    Uses ``NumPy`` to compute angles in a vectorized fashion.
    Sign of the torsion angle is also calculated.

    Uses Prof. Azevedo implementation:
    https://azevedolab.net/resources/dihedral_angle.pdf

    Example
    -------
    Given the sequential coords that represent a dummy molecule of
    four atoms:

    >>> xyz = numpy.array([
    >>>     [0.06360, -0.79573, 1.21644],
    >>>     [-0.47370, -0.10913, 0.77737],
    >>>     [-1.75288, -0.51877, 1.33236],
    >>>     [-2.29018, 0.16783, 0.89329],
    >>>     ])

    A1---A2
           \
            \
            A3---A4

    Calculates the torsion angle in A2-A3 that would place A4 in respect
    to the plane (A1, A2, A3).

    Likewise, for a chain of N atoms A1, ..., An, calculates the torsion
    angles in (A2, A3) to (An-2, An-1). (A1, A2) and (An-1, An) do not
    have torsion angles.

    If coords represent a protein backbone consisting of N, CA, and C
    atoms and starting at the N-terminal, the torsion angles are given
    by the following slices to the resulting array:

    - phi (N-CA), [2::3]
    - psi (CA-N), [::3]
    - omega (N-C), [1::3]

    Parameters
    ----------
    coords : numpy.ndarray of shape (N>=4, 3)
        Where `N` is the number of atoms, must be equal or above 4.

    Returns
    -------
    numpy.ndarray of shape (N - 3,)
        The torsion angles in radians.
        If you want to convert those to degrees just apply
        ``np.degrees`` to the returned result.
    """
    # requires
    assert coords.shape[0] > 3
    assert coords.shape[1] == 3

    crds = coords.T

    # Yes, I always write explicit array indices! :-)
    q_vecs = crds[:, 1:] - crds[:, :-1]
    cross = np.cross(q_vecs[:, :-1], q_vecs[:, 1:], axis=0)
    unitary = cross / np.linalg.norm(cross, axis=0)

    # components
    # u0 comes handy to define because it fits u1
    u0 = unitary[:, :-1]

    # u1 is the unitary cross products of the second plane
    # that is the unitary q2xq3, obviously applied to the whole chain
    u1 = unitary[:, 1:]

    # u3 is the unitary of the bonds that have a torsion representation,
    # those are all but the first and the last
    u3 = q_vecs[:, 1:-1] / np.linalg.norm(q_vecs[:, 1:-1], axis=0)

    # u2
    # there is no need to further select dimensions for u2, those have
    # been already sliced in u1 and u3.
    u2 = np.cross(u3, u1, axis=0)

    # calculating cos and sin of the torsion angle
    # here we need to use the .T and np.diagonal trick to achieve
    # broadcasting along the whole coords chain
    # np.matmul is preferred to np.dot in this case
    # https://numpy.org/doc/stable/reference/generated/numpy.matmul.html
    cos_theta = np.diagonal(np.matmul(u0.T, u1))
    sin_theta = np.diagonal(np.matmul(u0.T, u2))

    # torsion angles
    return -np.arctan2(sin_theta, cos_theta)


def torsion_set(p1, p2, p3, p4_vecs):
    """
    Calculate torsion angles of set torwards three initial vectors.

    Same implementation as `calc_torsion_angles` but A1, A2, and A3 are
    fixed and A4 is an array of vectors. The torsion angles are
    calculated for A4s with respect to the fixed A1, A2, and A3.

    Example
    -------
    >>> torsion_set(
        [x1, y1, z1],
        [x2, y2, z2],
        [x3, y3, z3],
        [
            [x4_1, y4_1, z4_1],
            [x4_2, y4_2, z4_2],
            [x4_3, y4_3, z4_3],
            [x4_4, y4_4, z4_4],
            ...
            ],
        )

    Returns
    -------
    numpy.ndarray of shape (N,)
        The anges in radians. Where N is the length of A4.
    """
    assert p4_vecs.shape[1] == 3
    q1 = p2 - p1
    q2 = p3 - p2
    q3s = p4_vecs - p3
    assert q3s.shape == p4_vecs.shape

    q1x2 = np.cross(q1, q2)
    assert q1x2.shape == (3,)
    q2x3s = np.cross(q2, q3s)
    assert q2x3s.shape == q3s.shape

    n1 = q1x2 / np.linalg.norm(q1x2)
    assert n1.shape == (3,)
    n2 = q2x3s / np.linalg.norm(q2x3s)
    assert n2.shape == p4_vecs.shape, f'{n2.shape}, {p4_vecs.shape}'

    u1 = n2
    u3 = q2 / np.linalg.norm(q2)
    u2 = np.cross(u3, u1)
    assert u2.shape == p4_vecs.shape

    cos_theta = np.dot(n1, u1.T)
    sen_theta = np.dot(n1, u2.T)

    theta = -np.arctan2(sen_theta, cos_theta)
    assert theta.size == p4_vecs.shape[0]
    return theta



"""Test libcalc."""
from math import sqrt

import MDAnalysis as mda
import numpy as np
import pytest

from taurenmd.libs import libcalc as lc

from . import toptest, trajtest


def test_mda_rmsd():
    """Test MDA RMSDS."""
    u = mda.Universe(toptest.str(), trajtest.str())
    result = lc.mda_rmsd(u)
    assert result.shape == (10,)
    assert result.dtype == np.float


def test_mda_rmsd_2():
    """Test MDA RMSDS."""
    u = mda.Universe(toptest.str(), trajtest.str())
    result = lc.mda_rmsd(u, frame_slice='::2')
    assert result.shape == (5,)
    assert result.dtype == np.float


def test_mda_rmsf():
    """Test MDA RMSF."""
    u = mda.Universe(toptest.str(), trajtest.str())
    CA = u.select_atoms('name CA')
    result = lc.mda_rmsf(CA, frame_slice=None)
    assert isinstance(result, np.ndarray)
    assert result.size == len(CA)
    assert result.dtype == np.float


@pytest.mark.parametrize(
    'p1,p2,p3,expected',
    [
        (
            np.array([0.0, 0.0, 0.0]),
            np.array([1.0, 0.0, 0.0]),
            np.array([0.0, 1.0, 0.0]),
            np.array([0.0, 0.0, 1.0])
            ),
        (
            np.array([0.0, 0.0, 0.0]),
            np.array([0.0, -1.0, 0.0]),
            np.array([0.0, 0.0, 1.0]),
            np.array([-1.0, 0.0, 0.0])
            ),
        ],
    )
def test_calc_plane_normal(p1, p2, p3, expected):
    """Test calc plane normal."""
    result = lc.calc_plane_normal(p1, p2, p3)
    assert np.all(np.equal(result, expected))


@pytest.mark.parametrize(
    'p1,p2,p3,expected',
    [
        (
            np.array([0.0, 0.0, 0.0]),
            np.array([1.0, 0.0, 0.0]),
            np.array([0.0, 1.0, 0.0]),
            (0.0, 0.0, 1.0, 0.0),
            ),
        (
            np.array([0.0, 0.0, 0.0]),
            np.array([0.0, -1.0, 0.0]),
            np.array([0.0, 0.0, 1.0]),
            (-1.0, 0.0, 0.0, 0.0)
            ),
        ],
    )
def test_calc_plane_eq(p1, p2, p3, expected):
    """Test calc plane normal."""
    result = lc.calc_plane_eq(p1, p2, p3)
    print(result)
    assert all(abs(a - b) < 0.00001 for a, b in zip(result, expected))


@pytest.mark.parametrize(
    'a1,b1,c1,a2,b2,c2,aunit,expected',
    [
        (0.0, 0.0, -1.0, 1.0, 0.0, 0.0, 'degrees', 90.0),
        (0.0, 0.0, -1.0, 1.0, 0.0, 0.0, 'radians', 1.5708),
        (0.0, 0.0, 1.0, 1.0, 0.0, 0.0, 'degrees', 90.0),
        (1.0, 0.0, 0.0, 1.0, 0.0, 0.0, 'degrees', 0.0),
        (1.0, 0.0, 0.0, 1.0, 0.0, 0.0, 'radians', 0.0),
        ],
    )
def test_calc_planes_angle(a1, b1, c1, a2, b2, c2, aunit, expected):
    """Test angle between planes."""
    result = lc.calc_planes_angle(a1, b1, c1, a2, b2, c2, aunit)
    assert abs(result - expected) < 0.00001


@pytest.mark.parametrize(
    'coords, expected',
    [
        (np.array([[1, 0, 0], [0, 0, 0], [0, 0, 1], [0, 1, 1]]), np.array([90], dtype=float)),  # noqa: E501
        (np.array([[1, 0, 0], [0, 0, 0], [0, 0, 1], [0, 1, 1], [1, 1, 1]]), np.array([90, -90], dtype=float)),  # noqa: E501
        (
            np.array((
                (1, 0, 0.000),
                (0, 0, 0.000),
                (0, 1, 0.000),
                (-1, 1, 0.000),
                (-1, 1, 1),
                (0, 1, 1),
                (0, 0, 0)
                )),
            np.array([-180, 90, -0, -45], dtype=float),
            ),
        (
            np.array((
                (1, 0, 0.000),
                (0, 0, 0.000),
                (0, 1, 0.000),
                (-1, 1, 0.000),
                (-1, 1, 1),
                )),
            np.array([-180, 90], dtype=float),
            ),
        ]
    )
def test_calculate_torsions(coords, expected):
    """Tests torsion calculation."""
    result = np.degrees(lc.calc_torsion_angles(coords))
    assert np.all(np.equal(result, expected))


@pytest.mark.parametrize(
    'p1,p2,p3,p4,expected',
    [
        (
            np.array([1, 0, 0]),
            np.array([0, 0, 0]),
            np.array([0, 1, 0]),
            np.array([
                [1, 1, 0],  # 0
                [0, 1, 1],
                [sqrt(1), 1, sqrt(1)],
                ]),
            np.array([0, -90, -45]),
            )
        ]
    )
def test_calculate_torsions_set(p1, p2, p3, p4, expected):
    """Tests torsion calculation."""
    result = np.degrees(lc.torsion_set(p1, p2, p3, p4))
    assert np.all(np.equal(result, expected))
